@@ -28,7 +28,6 @@
 
 #include <libgnomevfs/gnome-vfs-cancellation.h>
 #include <libgnomevfs/gnome-vfs-context.h>
-#include <libgnomevfs/gnome-vfs-i18n.h>
 #include <libgnomevfs/gnome-vfs-method.h>
 #include <libgnomevfs/gnome-vfs-mime.h>
 #include <libgnomevfs/gnome-vfs-module-shared.h>
@@ -38,23 +37,103 @@
 #include <libgnomevfs/gnome-vfs-monitor-private.h>
 #include <libgnomevfs/gnome-vfs-private-utils.h>
 #include <libgnomevfs/gnome-vfs-ops.h>
+#include <glib.h>
+#include <glib/gi18n-lib.h>
+#include <glib/gstdio.h>
+#ifndef G_OS_WIN32
 #include <dirent.h>
+#endif
 #include <errno.h>
 #include <fcntl.h>
-#include <glib/gstrfuncs.h>
-#include <glib/gutils.h>
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#ifndef G_OS_WIN32		/* We don't want the ftruncate() in mingw's unistd.h */
 #include <unistd.h>
+#endif
 #include <utime.h>
 #include <string.h>
 #ifdef HAVE_FAM
 #include <fam.h>
-#include <glib/giochannel.h>
+#endif
+
+#if HAVE_SYS_STATVFS_H
+#include <sys/statvfs.h>
+#endif
+#if HAVE_SYS_VFS_H
+#include <sys/vfs.h>
+#elif HAVE_SYS_MOUNT_H
+#include <sys/mount.h>
+#endif
+
+#ifdef G_OS_WIN32
+#include <windows.h>
+#endif
+
+#ifdef G_OS_WIN32
+#define DIR_SEPARATORS "/\\"
+#else
+#define DIR_SEPARATORS "/"
+#endif
+
+#if !GLIB_CHECK_VERSION (2,7,0)
+/* g_access lifted from GLib */
+#ifndef G_OS_WIN32
+#define g_access(filename, mode) access (filename, mode)
+#else
+static int
+g_access (const gchar *filename,
+	  int          mode)
+{
+#ifdef G_OS_WIN32
+  if (G_WIN32_HAVE_WIDECHAR_API ())
+    {
+      wchar_t *wfilename = g_utf8_to_utf16 (filename, -1, NULL, NULL, NULL);
+      int retval;
+      int save_errno;
+      
+      if (wfilename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = _waccess (wfilename, mode);
+      save_errno = errno;
+
+      g_free (wfilename);
+
+      errno = save_errno;
+      return retval;
+    }
+  else
+    {    
+      gchar *cp_filename = g_locale_from_utf8 (filename, -1, NULL, NULL, NULL);
+      int retval;
+      int save_errno;
+
+      if (cp_filename == NULL)
+	{
+	  errno = EINVAL;
+	  return -1;
+	}
+
+      retval = access (cp_filename, mode);
+      save_errno = errno;
+
+      g_free (cp_filename);
+
+      errno = save_errno;
+      return retval;
+    }
+#else
+  return access (filename, mode);
+#endif
+}
+#endif
 #endif
 
 #ifdef HAVE_FAM
@@ -97,7 +176,7 @@ GET_PATH_MAX (void)
 #ifdef HAVE_OPEN64
 #define OPEN open64
 #else
-#define OPEN open
+#define OPEN g_open
 #endif
 
 #if defined(HAVE_LSEEK64) && defined(HAVE_OFF64_T)
@@ -108,24 +187,62 @@ GET_PATH_MAX (void)
 #define OFF_T off_t
 #endif
 
+#ifdef G_OS_WIN32
+
+static int
+ftruncate (int     fd,
+	   guint64 size)
+{
+	/* FIXME: not threadsafe at all! */
+	LARGE_INTEGER origpos;
+	int retval = -1;
+
+	origpos.QuadPart = 0;
+	origpos.u.LowPart = SetFilePointer ((HANDLE) _get_osfhandle (fd), 0,
+					    &origpos.u.HighPart, FILE_CURRENT);
+	if (origpos.u.LowPart != INVALID_SET_FILE_POINTER) {
+		LARGE_INTEGER newpos;
+
+		newpos.QuadPart = size;
+		if (SetFilePointer ((HANDLE) _get_osfhandle (fd), newpos.u.LowPart,
+				    &newpos.u.HighPart, FILE_BEGIN) != INVALID_SET_FILE_POINTER &&
+		    SetEndOfFile ((HANDLE) _get_osfhandle (fd)) &&
+		    SetFilePointer ((HANDLE) _get_osfhandle (fd), origpos.u.LowPart,
+				    &origpos.u.HighPart, FILE_BEGIN) != INVALID_SET_FILE_POINTER)
+			retval = 0;
+	}
+
+	if (retval == -1)
+		errno = EIO;
+
+	return retval;
+}
+
+#endif
 
 static gchar *
 get_path_from_uri (GnomeVFSURI const *uri)
 {
 	gchar *path;
 
-	path = gnome_vfs_unescape_string (uri->text, 
-		G_DIR_SEPARATOR_S);
+	path = gnome_vfs_unescape_string (uri->text, DIR_SEPARATORS);
 		
 	if (path == NULL) {
 		return NULL;
 	}
 
-	if (path[0] != G_DIR_SEPARATOR) {
+	if (!g_path_is_absolute (path)) {
 		g_free (path);
 		return NULL;
 	}
-
+#ifdef G_OS_WIN32
+	/* Drop slash in front of drive letter */
+	if (path[0] == '/' && g_ascii_isalpha (path[1]) && path[2] == ':') {
+		gchar *tem = path;
+		path = g_strdup (path + 1);
+		g_free (tem);
+	}
+#endif
 	return path;
 }
 
@@ -135,7 +252,7 @@ get_base_from_uri (GnomeVFSURI const *uri)
 	gchar *escaped_base, *base;
 
 	escaped_base = gnome_vfs_uri_extract_short_path_name (uri);
-	base = gnome_vfs_unescape_string (escaped_base, G_DIR_SEPARATOR_S);
+	base = gnome_vfs_unescape_string (escaped_base, DIR_SEPARATORS);
 	g_free (escaped_base);
 	return base;
 }
@@ -192,6 +309,9 @@ do_open (GnomeVFSMethod *method,
 		else
 			return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
 	}
+#ifdef G_OS_WIN32
+	unix_mode |= _O_BINARY;
+#endif
 
 	if ((mode & GNOME_VFS_OPEN_TRUNCATE) ||
 	    (!(mode & GNOME_VFS_OPEN_RANDOM) && (mode & GNOME_VFS_OPEN_WRITE)))
@@ -202,7 +322,7 @@ do_open (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_INVALID_URI;
 
 	do
-		fd = OPEN (file_name, unix_mode);
+		fd = OPEN (file_name, unix_mode, 0);
 	while (fd == -1
 	       && errno == EINTR
 	       && ! gnome_vfs_context_check_cancellation (context));
@@ -252,6 +372,9 @@ do_create (GnomeVFSMethod *method,
 
 	unix_mode = O_CREAT | O_TRUNC;
 	
+#ifdef G_OS_WIN32
+	unix_mode |= _O_BINARY;
+#endif
 	if (!(mode & GNOME_VFS_OPEN_WRITE))
 		return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
 
@@ -441,7 +564,7 @@ do_seek (GnomeVFSMethod *method,
 static GnomeVFSResult
 do_tell (GnomeVFSMethod *method,
 	 GnomeVFSMethodHandle *method_handle,
-	 GnomeVFSFileOffset *offset_return)
+	 GnomeVFSFileSize *offset_return)
 {
 	FileHandle *file_handle;
 	OFF_T offset;
@@ -494,6 +617,7 @@ do_truncate (GnomeVFSMethod *method,
 	     GnomeVFSFileSize where,
 	     GnomeVFSContext *context)
 {
+#ifndef G_OS_WIN32
 	gchar *path;
 
 	path = get_path_from_uri (uri);
@@ -515,22 +639,32 @@ do_truncate (GnomeVFSMethod *method,
 			return GNOME_VFS_ERROR_GENERIC;
 		}
 	}
+#else
+	g_warning ("Not implemented: file::do_truncate()");
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+#endif
 }
 
 typedef struct {
 	GnomeVFSURI *uri;
-	DIR *dir;
 	GnomeVFSFileInfoOptions options;
-
+#ifndef G_OS_WIN32
+	DIR *dir;
 	struct dirent *current_entry;
-
+#else
+	GDir *dir;
+#endif
 	gchar *name_buffer;
 	gchar *name_ptr;
 } DirectoryHandle;
 
 static DirectoryHandle *
 directory_handle_new (GnomeVFSURI *uri,
+#ifndef G_OS_WIN32
 		      DIR *dir,
+#else
+		      GDir *dir,
+#endif
 		      GnomeVFSFileInfoOptions options)
 {
 	DirectoryHandle *result;
@@ -542,8 +676,10 @@ directory_handle_new (GnomeVFSURI *uri,
 	result->uri = gnome_vfs_uri_ref (uri);
 	result->dir = dir;
 
+#ifndef G_OS_WIN32
 	/* Reserve extra space for readdir_r, see man page */
 	result->current_entry = g_malloc (sizeof (struct dirent) + GET_PATH_MAX() + 1);
+#endif
 
 	full_name = get_path_from_uri (uri);
 	g_assert (full_name != NULL); /* already done by caller */
@@ -552,8 +688,8 @@ directory_handle_new (GnomeVFSURI *uri,
 	result->name_buffer = g_malloc (full_name_len + GET_PATH_MAX () + 2);
 	memcpy (result->name_buffer, full_name, full_name_len);
 	
-	if (full_name_len > 0 && full_name[full_name_len - 1] != '/')
-		result->name_buffer[full_name_len++] = '/';
+	if (full_name_len > 0 && !G_IS_DIR_SEPARATOR (full_name[full_name_len - 1]))
+		result->name_buffer[full_name_len++] = G_DIR_SEPARATOR;
 
 	result->name_ptr = result->name_buffer + full_name_len;
 
@@ -569,7 +705,9 @@ directory_handle_destroy (DirectoryHandle *directory_handle)
 {
 	gnome_vfs_uri_unref (directory_handle->uri);
 	g_free (directory_handle->name_buffer);
+#ifndef G_OS_WIN32
 	g_free (directory_handle->current_entry);
+#endif
 	g_free (directory_handle);
 }
 
@@ -607,6 +745,7 @@ get_mime_type (GnomeVFSFileInfo *info,
 	info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
 }
 
+#ifndef G_OS_WIN32
 static gchar *
 read_link (const gchar *full_name)
 {
@@ -632,6 +771,7 @@ read_link (const gchar *full_name)
 		buffer = g_realloc (buffer, size);
 	}
 }
+#endif
 
 static void
 get_access_info (GnomeVFSFileInfo *file_info,
@@ -641,17 +781,24 @@ get_access_info (GnomeVFSFileInfo *file_info,
       * want to set valid_fields if something bad happened during one
       * of the access calls
       */
-     if (access (full_name, R_OK) == 0) {
+     if (g_access (full_name, R_OK) == 0) {
              file_info->permissions |= GNOME_VFS_PERM_ACCESS_READABLE;
      }
 
-     if (access (full_name, W_OK) == 0) {
+     if (g_access (full_name, W_OK) == 0) {
              file_info->permissions |= GNOME_VFS_PERM_ACCESS_WRITABLE;
      }
 
-     if (access (full_name, X_OK) == 0) {
+#ifdef G_OS_WIN32
+     if (g_file_test (full_name, G_FILE_TEST_IS_EXECUTABLE)) {
              file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
      }
+#else 
+     if (g_access (full_name, X_OK) == 0) {
+             file_info->permissions |= GNOME_VFS_PERM_ACCESS_EXECUTABLE;
+     }
+#endif 
+
      file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_ACCESS;
 }
 
@@ -663,12 +810,14 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 {
 	struct stat statbuf;
 	gboolean followed_symlink;
+#ifndef G_OS_WIN32
 	gboolean is_symlink;
-	gboolean recursive;
 	char *link_file_path;
 	char *symlink_name;
 	char *symlink_dir;
 	char *newpath;
+#endif
+	gboolean recursive;
 	
 	followed_symlink = FALSE;
 	
@@ -680,14 +829,16 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 		statptr = &statbuf;
 	}
 
-	if (lstat (full_name, statptr) != 0) {
+	if (g_lstat (full_name, statptr) != 0) {
 		return gnome_vfs_result_from_errno ();
 	}
 
+
+#ifndef G_OS_WIN32
 	is_symlink = S_ISLNK (statptr->st_mode);
 
 	if ((options & GNOME_VFS_FILE_INFO_FOLLOW_LINKS) && is_symlink) {
-		if (stat (full_name, statptr) != 0) {
+		if (g_stat (full_name, statptr) != 0) {
 			if (errno == ELOOP) {
 				recursive = TRUE;
 			}
@@ -695,16 +846,17 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 			/* It's a broken symlink, revert to the lstat. This is sub-optimal but
 			 * acceptable because it's not a common case.
 			 */
-			if (lstat (full_name, statptr) != 0) {
+			if (g_lstat (full_name, statptr) != 0) {
 				return gnome_vfs_result_from_errno ();
 			}
 		}
 		GNOME_VFS_FILE_INFO_SET_SYMLINK (file_info, TRUE);
 		followed_symlink = TRUE;
 	}
-
+#endif
 	gnome_vfs_stat_to_file_info (file_info, statptr);
 
+#ifndef G_OS_WIN32
 	if (is_symlink) {
 		symlink_name = NULL;
 		link_file_path = g_strdup (full_name);
@@ -741,7 +893,7 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 			                /* if we had an earlier ELOOP, don't get in an infinite loop here */
 			        || recursive
 					/* we don't care to follow links */
-				|| lstat (symlink_name, statptr) != 0
+				|| g_lstat (symlink_name, statptr) != 0
 					/* we can't make out where this points to */
 				|| !S_ISLNK (statptr->st_mode)) {
 					/* the next level is not a link */
@@ -754,7 +906,7 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 
 		file_info->symlink_name = symlink_name;
 	}
-
+#endif
 	return GNOME_VFS_OK;
 }
 
@@ -789,13 +941,21 @@ do_open_directory (GnomeVFSMethod *method,
 		   GnomeVFSContext *context)
 {
 	gchar *directory_name;
+#ifndef G_OS_WIN32
 	DIR *dir;
+#else
+	GDir *dir;
+#endif
 
 	directory_name = get_path_from_uri (uri);
 	if (directory_name == NULL)
 		return GNOME_VFS_ERROR_INVALID_URI;
 
+#ifndef G_OS_WIN32
 	dir = opendir (directory_name);
+#else
+	dir = g_dir_open (directory_name, 0, NULL);
+#endif
 	g_free (directory_name);
 	if (dir == NULL)
 		return gnome_vfs_result_from_errno ();
@@ -816,7 +976,11 @@ do_close_directory (GnomeVFSMethod *method,
 
 	directory_handle = (DirectoryHandle *) method_handle;
 
+#ifndef G_OS_WIN32
 	closedir (directory_handle->dir);
+#else
+	g_dir_close (directory_handle->dir);
+#endif
 
 	directory_handle_destroy (directory_handle);
 
@@ -833,7 +997,11 @@ do_read_directory (GnomeVFSMethod *method,
 		   GnomeVFSFileInfo *file_info,
 		   GnomeVFSContext *context)
 {
+#ifndef G_OS_WIN32
 	struct dirent *result;
+#else
+	const gchar *result;
+#endif
 	struct stat statbuf;
 	gchar *full_name;
 	DirectoryHandle *handle;
@@ -854,16 +1022,22 @@ do_read_directory (GnomeVFSMethod *method,
 #else
 	G_LOCK (readdir);
 	errno = 0;
+#ifndef G_OS_WIN32
 	result = readdir (handle->dir);
+#else
+	result = g_dir_read_name (handle->dir);
+#endif
 
 	if (result == NULL && errno != 0) {
 		GnomeVFSResult ret = gnome_vfs_result_from_errno ();
 		G_UNLOCK (readdir);
 		return ret;
 	}
+#ifndef G_OS_WIN32
 	if (result != NULL) {
 		memcpy (handle->current_entry, result, sizeof (struct dirent));
 	}
+#endif
 	G_UNLOCK (readdir);
 #endif
 	
@@ -871,9 +1045,13 @@ do_read_directory (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_EOF;
 	}
 
+#ifndef G_OS_WIN32
 	file_info->name = g_strdup (result->d_name);
-
 	strcpy (handle->name_ptr, result->d_name);
+#else
+	file_info->name = g_strdup (result);
+	strcpy (handle->name_ptr, result);
+#endif
 	full_name = handle->name_buffer;
 
 	if (get_stat_info (file_info, full_name, handle->options, &statbuf) != GNOME_VFS_OK) {
@@ -996,7 +1174,7 @@ do_is_local (GnomeVFSMethod *method,
 
 	if (local == NULL) {
 		struct stat statbuf;
-		if (stat (path, &statbuf) == 0) {
+		if (g_stat (path, &statbuf) == 0) {
 			char *type = filesystem_type (path, path, &statbuf);
 			gboolean is_local = ((strcmp (type, "nfs") != 0) && 
 					     (strcmp (type, "afs") != 0) &&
@@ -1025,7 +1203,7 @@ do_make_directory (GnomeVFSMethod *method,
 	if (full_name == NULL)
 		return GNOME_VFS_ERROR_INVALID_URI;
 
-	retval = mkdir (full_name, perm);
+	retval = g_mkdir (full_name, perm);
 
 	g_free (full_name);
 
@@ -1048,7 +1226,7 @@ do_remove_directory (GnomeVFSMethod *method,
 	if (full_name == NULL)
 		return GNOME_VFS_ERROR_INVALID_URI;
 
-	retval = rmdir (full_name);
+	retval = g_rmdir (full_name);
 
 	g_free (full_name);
 
@@ -1082,14 +1260,14 @@ mkdir_recursive (const char *path, int permission_bits)
 			if (!*dir_separator_scanner) {
 				break;
 			}	
-			if (*dir_separator_scanner == G_DIR_SEPARATOR) {
+			if (G_IS_DIR_SEPARATOR (*dir_separator_scanner)) {
 				break;
 			}
 		}
 		if (dir_separator_scanner - path > 0) {
 			current_path = g_strndup (path, dir_separator_scanner - path);
-			mkdir (current_path, permission_bits);
-			if (stat (current_path, &stat_buffer) != 0) {
+			g_mkdir (current_path, permission_bits);
+			if (g_stat (current_path, &stat_buffer) != 0) {
 				/* we failed to create a directory and it wasn't there already;
 				 * bail
 				 */
@@ -1109,23 +1287,24 @@ mkdir_recursive (const char *path, int permission_bits)
 static char *
 append_to_path (const char *path, const char *name)
 {
-	return g_strconcat (path, G_DIR_SEPARATOR_S, name, NULL);
+	return g_build_filename (path, name, NULL);
 }
 
 static char *
 append_trash_path (const char *path)
 {	
-	/* When creating trash outside of /home/pavel, create it in the form:
-	 * .Trash-pavel to allow sharing the name space for several users.
-	 * Treat "/" specially to avoid creating non-canonical "//foo" path.
-	 */
-	if (strcmp (path, "/") == 0) {
-		return g_strconcat (path, TRASH_DIRECTORY_NAME_BASE,
-		"-", g_get_user_name (), NULL);
-	} else {
-		return g_strconcat (path, G_DIR_SEPARATOR_S, TRASH_DIRECTORY_NAME_BASE,
-			"-", g_get_user_name (), NULL);
-	}
+	char *per_user_part; 
+	char *retval;
+	
+	per_user_part = g_strconcat (TRASH_DIRECTORY_NAME_BASE "-", 
+			             g_get_user_name (), 
+				     NULL);
+
+	retval = g_build_filename (path, per_user_part, NULL);
+	
+	g_free (per_user_part);
+
+	return retval;
 }
 
 static char *
@@ -1139,7 +1318,7 @@ find_trash_in_hierarchy (const char *start_dir, dev_t near_device_id, GnomeVFSCo
 
 	/* check if there is a trash in this directory */
 	trash_path = append_trash_path (start_dir);
-	if (lstat (trash_path, &stat_buffer) == 0 && S_ISDIR (stat_buffer.st_mode)) {
+	if (g_lstat (trash_path, &stat_buffer) == 0 && S_ISDIR (stat_buffer.st_mode)) {
 		/* found it, we are done */
 		g_assert (near_device_id == stat_buffer.st_dev);
 		return trash_path;
@@ -1221,7 +1400,7 @@ find_disk_top_directory (const char *item_on_disk,
 		}
 		
 		*last_slash = '\0';
-		if (stat (disk_top_directory, &stat_buffer) < 0
+		if (g_stat (disk_top_directory, &stat_buffer) < 0
 			|| stat_buffer.st_dev != near_device_id) {
 			/* we ran past the root of the disk we are exploring */
 			g_free (disk_top_directory);
@@ -1265,7 +1444,7 @@ save_trash_entry_cache (void)
 		return;
 	}
 
-	cache_file = open (cache_file_path, O_CREAT | O_TRUNC | O_RDWR, 0666);
+	cache_file = g_open (cache_file_path, O_CREAT | O_TRUNC | O_RDWR, 0666);
 	if (cache_file < 0) {
 		g_warning ("failed to create trash item cache file");
 		return;
@@ -1396,9 +1575,11 @@ read_saved_cached_trash_entries (void)
 	cached_trash_directories = NULL;
 
 	/* read in the entries from disk */
-	cache_file_path = g_strconcat (g_get_home_dir (), G_DIR_SEPARATOR_S,
-		TRASH_ENTRY_CACHE_PARENT, G_DIR_SEPARATOR_S, TRASH_ENTRY_CACHE_NAME, NULL);
-	cache_file = fopen (cache_file_path, "r");
+	cache_file_path = g_build_filename (g_get_home_dir (),
+					    TRASH_ENTRY_CACHE_PARENT,
+					    TRASH_ENTRY_CACHE_NAME,
+					    NULL);
+	cache_file = g_fopen (cache_file_path, "r");
 
 	if (cache_file != NULL) {
 		removed_item = FALSE;
@@ -1411,13 +1592,13 @@ read_saved_cached_trash_entries (void)
 			trash_path = NULL;
 			if (sscanf (buffer, "%s %s", escaped_mount_point, escaped_trash_path) == 2) {
 				/* the paths are saved in escaped form */
-				trash_path = gnome_vfs_unescape_string (escaped_trash_path, "/");
-				mount_point = gnome_vfs_unescape_string (escaped_mount_point, "/"); 
+				trash_path = gnome_vfs_unescape_string (escaped_trash_path, DIR_SEPARATORS);
+				mount_point = gnome_vfs_unescape_string (escaped_mount_point, DIR_SEPARATORS); 
 
 				if (trash_path != NULL 
 					&& mount_point != NULL
-					&& (strcmp (trash_path, NON_EXISTENT_TRASH_ENTRY) != 0 && lstat (trash_path, &stat_buffer) == 0)
-					&& stat (mount_point, &stat_buffer) == 0) {
+					&& (strcmp (trash_path, NON_EXISTENT_TRASH_ENTRY) != 0 && g_lstat (trash_path, &stat_buffer) == 0)
+					&& g_stat (mount_point, &stat_buffer) == 0) {
 					/* We know the trash exist and we checked that it's really
 					 * there - this is a good entry, copy it into the local cache.
 					 * We don't want to rely on old non-existing trash entries, as they
@@ -1459,7 +1640,7 @@ static gboolean
 cached_trash_entry_exists (const TrashDirectoryCachedItem *entry)
 {
 	struct stat stat_buffer;
-	return lstat (entry->path, &stat_buffer) == 0;
+	return g_lstat (entry->path, &stat_buffer) == 0;
 }
 
 /* Search through the local cache looking for an entry that matches a given
@@ -1673,7 +1854,7 @@ do_find_directory (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_CANCELLED;
 	}
 
-	retval = lstat (full_name_near, &near_item_stat);
+	retval = g_lstat (full_name_near, &near_item_stat);
 	if (retval != 0) {
 		g_free (full_name_near);
 		return gnome_vfs_result_from_errno ();
@@ -1684,7 +1865,7 @@ do_find_directory (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_CANCELLED;
 	}
 	
-	retval = stat (home_directory, &home_volume_stat);
+	retval = g_stat (home_directory, &home_volume_stat);
 	if (retval != 0) {
 		g_free (full_name_near);
 		return gnome_vfs_result_from_errno ();
@@ -1743,11 +1924,11 @@ do_find_directory (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
 	}
 
-	if (create_if_needed && access (target_directory_path, F_OK) != 0) {
+	if (create_if_needed && g_access (target_directory_path, F_OK) != 0) {
 		mkdir_recursive (target_directory_path, permissions);
 	}
 
-	if (access (target_directory_path, F_OK) != 0) {
+	if (g_access (target_directory_path, F_OK) != 0) {
 		g_free (target_directory_path);
 		return GNOME_VFS_ERROR_NOT_FOUND;
 	}
@@ -1773,7 +1954,7 @@ rename_helper (const gchar *old_full_name,
 	GnomeVFSHandle *temp_handle;
 	GnomeVFSResult result;
 
-	retval = stat (new_full_name, &statbuf);
+	retval = g_stat (new_full_name, &statbuf);
 	if (retval == 0) {
 		/* Special case for files on case insensitive (vfat) filesystems:
 		 * If the old and the new name only differ by case,
@@ -1789,17 +1970,17 @@ rename_helper (const gchar *old_full_name,
 			if (result != GNOME_VFS_OK)
 				return result;
 			gnome_vfs_close (temp_handle);
-			unlink (temp_name);
+			g_unlink (temp_name);
 			
-			retval = rename (old_full_name, temp_name);
+			retval = g_rename (old_full_name, temp_name);
 			if (retval == 0) {
-				if (stat (new_full_name, &statbuf) != 0 
-				    && rename (temp_name, new_full_name) == 0) {
+				if (g_stat (new_full_name, &statbuf) != 0 
+				    && g_rename (temp_name, new_full_name) == 0) {
 					/* Success */
 					return GNOME_VFS_OK;
 				}
 				/* Revert the filename back to original */ 
-				retval = rename (temp_name, old_full_name);
+				retval = g_rename (temp_name, old_full_name);
 				if (retval == 0) {
 					return GNOME_VFS_ERROR_FILE_EXISTS;
 				}
@@ -1820,7 +2001,7 @@ rename_helper (const gchar *old_full_name,
 	if (gnome_vfs_context_check_cancellation (context))
 		return GNOME_VFS_ERROR_CANCELLED;
 
-	retval = rename (old_full_name, new_full_name);
+	retval = g_rename (old_full_name, new_full_name);
 
 	/* FIXME bugzilla.eazel.com 1186: The following assumes that,
 	 * if `new_uri' and `old_uri' are on different file systems,
@@ -1837,7 +2018,7 @@ rename_helper (const gchar *old_full_name,
 		if (S_ISDIR (statbuf.st_mode)) {
 			if (gnome_vfs_context_check_cancellation (context))
 				return GNOME_VFS_ERROR_CANCELLED;
-			retval = rmdir (new_full_name);
+			retval = g_rmdir (new_full_name);
 			if (retval != 0) {
 				return gnome_vfs_result_from_errno ();
 			}
@@ -1845,7 +2026,7 @@ rename_helper (const gchar *old_full_name,
 			if (gnome_vfs_context_check_cancellation (context))
 				return GNOME_VFS_ERROR_CANCELLED;
 
-			retval = rename (old_full_name, new_full_name);
+			retval = g_rename (old_full_name, new_full_name);
 		}
 	}
 
@@ -1899,7 +2080,7 @@ do_unlink (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_INVALID_URI;
 	}
 
-	retval = unlink (full_name);
+	retval = g_unlink (full_name);
 
 	g_free (full_name);
 
@@ -1916,6 +2097,7 @@ do_create_symbolic_link (GnomeVFSMethod *method,
 			 const char *target_reference,
 			 GnomeVFSContext *context)
 {
+#ifndef G_OS_WIN32
 	const char *link_scheme, *target_scheme;
 	char *link_full_name, *target_full_name;
 	GnomeVFSResult result;
@@ -1967,6 +2149,9 @@ do_create_symbolic_link (GnomeVFSMethod *method,
 	gnome_vfs_uri_unref (target_uri);
 
 	return result;
+#else
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+#endif
 }
 
 /* When checking whether two locations are on the same file system, we are
@@ -1988,7 +2173,7 @@ do_check_same_fs (GnomeVFSMethod *method,
 	gint retval;
 
 	full_name_source = get_path_from_uri (source_uri);
-	retval = lstat (full_name_source, &s_source);
+	retval = g_lstat (full_name_source, &s_source);
 	g_free (full_name_source);
 
 	if (retval != 0)
@@ -1998,7 +2183,7 @@ do_check_same_fs (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_CANCELLED;
  
 	full_name_target = get_path_from_uri (target_uri);
-	retval = stat (full_name_target, &s_target);
+	retval = g_stat (full_name_target, &s_target);
 	g_free (full_name_target);
 
 	if (retval != 0)
@@ -2028,20 +2213,11 @@ do_set_file_info (GnomeVFSMethod *method,
 		gchar *new_name;
 
 		encoded_dir = gnome_vfs_uri_extract_dirname (uri);
-		dir = gnome_vfs_unescape_string (encoded_dir, G_DIR_SEPARATOR_S);
+		dir = gnome_vfs_unescape_string (encoded_dir, DIR_SEPARATORS);
 		g_free (encoded_dir);
 		g_assert (dir != NULL);
 
-		/* FIXME bugzilla.eazel.com 645: This needs to return
-		 * an error for incoming names with "/" characters in
-		 * them, instead of moving the file.
-		 */
-
-		if (dir[strlen(dir) - 1] != '/') {
-			new_name = g_strconcat (dir, "/", info->name, NULL);
-		} else {
-			new_name = g_strconcat (dir, info->name, NULL);
-		}
+		new_name = g_build_filename (dir, info->name, NULL);
 
 		result = rename_helper (full_name, new_name, FALSE, context);
 
@@ -2072,12 +2248,15 @@ do_set_file_info (GnomeVFSMethod *method,
 	}
 
 	if (mask & GNOME_VFS_SET_FILE_INFO_OWNER) {
+#ifndef G_OS_WIN32
 		if (chown (full_name, info->uid, info->gid) != 0) {
 			g_free (full_name);
 			return gnome_vfs_result_from_errno ();
 		}
+#else
+		g_warning ("Not implemented: GNOME_VFS_SET_FILE_INFO_OWNER");
+#endif
 	}
-
 	if (gnome_vfs_context_check_cancellation (context)) {
 		g_free (full_name);
 		return GNOME_VFS_ERROR_CANCELLED;
@@ -2330,6 +2509,93 @@ do_file_control (GnomeVFSMethod *method,
 	return GNOME_VFS_ERROR_NOT_SUPPORTED;
 }
 
+static GnomeVFSResult
+do_get_volume_free_space (GnomeVFSMethod *method,
+			  const GnomeVFSURI *uri,
+			  GnomeVFSFileSize *free_space)
+{
+#ifndef G_OS_WIN32
+	GnomeVFSFileSize free_blocks, block_size;
+       	int statfs_result;
+	const char *path;
+	char *unescaped_path;
+
+#if HAVE_STATVFS
+	struct statvfs statfs_buffer;
+#else
+	struct statfs statfs_buffer;
+#endif
+
+ 	*free_space = 0;
+
+	path = gnome_vfs_uri_get_path (uri);
+	if (path == NULL || !_gnome_vfs_istr_has_prefix (path, "/")) {
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
+
+	unescaped_path = gnome_vfs_unescape_string (path, G_DIR_SEPARATOR_S);
+	
+#if HAVE_STATVFS
+	statfs_result = statvfs (unescaped_path, &statfs_buffer);
+	block_size = statfs_buffer.f_frsize; 
+#else
+#if STATFS_ARGS == 2
+	statfs_result = statfs (unescaped_path, &statfs_buffer);
+#elif STATFS_ARGS == 4
+	statfs_result = statfs (unescaped_path, &statfs_buffer,
+				sizeof (statfs_buffer), 0);
+#endif
+	block_size = statfs_buffer.f_bsize; 
+#endif  
+
+	if (statfs_result != 0) {
+		g_free (unescaped_path);
+		return gnome_vfs_result_from_errno ();
+	}
+
+
+/* CF: I assume ncpfs is linux specific, if you are on a non-linux platform
+ * where ncpfs is available, please file a bug about it on bugzilla.gnome.org
+ * (2004-03-08)
+ */
+#if defined(__linux__)
+	/* ncpfs does not know the amount of available and free space */
+	if (statfs_buffer.f_bavail == 0 && statfs_buffer.f_bfree == 0) {
+		/* statvfs does not contain an f_type field, we try again
+		 * with statfs.
+		 */
+		struct statfs statfs_buffer2;
+		statfs_result = statfs (unescaped_path, &statfs_buffer2);
+		g_free (unescaped_path);
+
+		if (statfs_result != 0) {
+			return gnome_vfs_result_from_errno ();
+		}
+		
+		/* linux/ncp_fs.h: NCP_SUPER_MAGIC == 0x564c */
+		if (statfs_buffer2.f_type == 0x564c)
+		{
+			return GNOME_VFS_ERROR_NOT_SUPPORTED;
+		}
+	} else {
+		/* everything is copacetic... free the unescaped path */
+		g_free (unescaped_path);
+	}
+#else 	
+	g_free (unescaped_path);
+#endif
+	free_blocks = statfs_buffer.f_bavail;
+
+	*free_space = block_size * free_blocks;
+	
+	return GNOME_VFS_OK;
+#else // G_OS_WIN32
+	g_warning ("Not implemented for WIN32: file-method.c:do_get_volume_free_space()");
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+#endif
+}
+
+
 static GnomeVFSMethod method = {
 	sizeof (GnomeVFSMethod),
 	do_open,
@@ -2358,7 +2624,8 @@ static GnomeVFSMethod method = {
 	do_monitor_add,
 	do_monitor_cancel,
 	do_file_control,
-	do_forget_cache
+	do_forget_cache,
+	do_get_volume_free_space
 };
 
 GnomeVFSMethod *
